"""NamedTuple and (basically) setters/getters used in secretctl cli"""
import json
import sys
from typing import NamedTuple
import boto3
from botocore.exceptions import ClientError

class Secret(NamedTuple):
    """immutable secret class for working with AWS Secrets Manager"""
    path: str
    value: str = None
    description: str = None
    tags: list = None
    versions: list = None

Secret.__doc__ = "Secret: namedtuple for Secret object in secretctl"
Secret.path.__doc__ = "[Name,SecretId] path/to/secret/key"
Secret.value.__doc__ = "[SecretString] secret value (SecretBinary not yet supported)"
Secret.description.__doc__ = "[Description] a description of the secret"
Secret.tags.__doc__ = "[Tags] tags in form of [{ \"Key\": \"<tag name>\", \"Value\": \"<tag value>\"}]"
Secret.versions.__doc__ = "[Versions] from list_secret_version_ids()"

_session = boto3.session.Session().client(service_name='secretsmanager', region_name='us-east-1')

def create_secret(path, value, description=None, tags=None):
    """create path:value in AWS Secrets Manager"""
    secret_kwargs = {}
    try:
        secret_kwargs['Name'] = path
        secret_kwargs['SecretString'] = value
        if description:
            secret_kwargs['Description'] = description
        if tags:
            secret_kwargs['Tags'] = tags
        resp = _session.create_secret(**secret_kwargs)
        return Secret(path=resp['Name'])
    except ClientError as e:
        # use secretctl update to change existing secrets
        print(e)
        sys.exit(1)

def get_secret(path):
    #pylint: disable=R1260
    """get path from from AWS Secrets Manager"""
    secret_kwargs = {}
    try:
        secret_kwargs['path'] = path
        resp = _session.describe_secret(SecretId=path)
        if 'Description' in resp:
            secret_kwargs['description'] = resp['Description']
        if 'Tags' in resp:
            secret_kwargs['tags'] = resp['Tags']
        resp = _session.get_secret_value(SecretId=path)
        if 'SecretString' in resp:
            secret_kwargs['value'] = json.loads(resp['SecretString'])
        secret_kwargs['versions'] = len(_session.list_secret_version_ids(SecretId=path)['Versions'])
        return Secret(**secret_kwargs)
    except ClientError as e:
        print(e)
        sys.exit(1)

def update_secret(path, value, description=None):
    """modify secret (causing new version), update description and/or tags"""
    secret_kwargs = {}
    try:
        resp = _session.describe_secret(SecretId=path)
        secret_kwargs['SecretId'] = path
        secret_kwargs['SecretString'] = value
        resp = _session.put_secret_value(**secret_kwargs)
        if description:
            secret_kwargs['Description'] = description
            resp = _session.update_secret(**secret_kwargs)
        return Secret(path=resp['Name'])
    except ClientError as e:
        print(e)
        sys.exit(1)

# removed from code coverage pending moto support for mocking secretsmanager tags
def tag_secret(path, tags=None):  # pragma: no cover
    """add tags to secret"""
    secret_kwargs = {}
    secret_kwargs['SecretId'] = path
    secret_kwargs['Tags'] = tags
    try:
        _session.tag_resource(**secret_kwargs)
        return Secret(path=path)
    except ClientError as e:
        print(e)
        sys.exit(1)

# removed from code coverage pending moto support for mocking secretsmanager tags
def untag_secret(path, tags=None):  # pragma: no cover
    """remove tags from secret"""
    secret_kwargs = {}
    secret_kwargs['SecretId'] = path
    secret_kwargs['TagKeys'] = tags
    try:
        _session.untag_resource(**secret_kwargs)
        return Secret(path=path)
    except ClientError as e:
        print(e)
        sys.exit(1)

def list_secrets():
    """list all secrets"""
    _secrets = []
    paginator = _session.get_paginator('list_secrets')
    page_iterator = paginator.paginate()
    for page in page_iterator:
        _secrets += page['SecretList']
    return _secrets

def delete_secret(path, recovery=7):
    """delete target secret"""
    try:
        resp = _session.delete_secret(SecretId=path, RecoveryWindowInDays=recovery)
        return resp
    except ClientError as e:
        print(e)
        sys.exit(1)



"""cli output formatters"""
import json
from colorama import init, Fore
from secretctl.validators import json_to_tags
from secretctl.tuples import Secret
init(autoreset=True)

MAX_DESC_WIDTH = 42
MIN_DESC_WIDTH = 18

# print formatted results of read to stdout
def print_read(secret, quiet=False, info=False):
    """format secretctl read response"""
    options = {True: print_read_quiet, False: print_read_normal}
    options[quiet](secret, info)

def print_read_quiet(secret, _=False):
    """read quiet"""
    for _, value in secret.value.items():
        print(value)

def print_read_normal(secret, info=False):
    """format normal output from read"""
    col_path = len(secret.path) + 3
    print(Fore.YELLOW + "{:{wid}} {:<9} {}".format('Path/Key', 'Version', 'Value', wid=col_path))
    # output full json when the secret is multiple key:value pairs
    _value = json.dumps(secret.value) if len(secret.value) >= 2 else next(iter(secret.value.values()))
    print("{:{wid}} {:<9} {}".format(secret.path, secret.versions, _value, wid=col_path))
    if info:
        _desc = secret.description if secret.description else 'No description'
        _tags = json_to_tags(secret.tags) if secret.tags else 'No tags'
        print("{:{wid}} {}".format('*Desc', _desc, wid=col_path+10))
        print("{:{wid}} {}".format('*Tags', _tags, wid=col_path+10))

# print formatted results of list to stdout
def print_list(secrets):
    """format secretctl list response"""
    line = "{:{path_wid}} {:{desc_wid}} {}"
    path_col = max(len(secret.path) for secret in secrets) + 3
    desc_col = desc_col_length(secrets)

    print(Fore.YELLOW + line.format('Path/Key', 'Description', 'Tags', path_wid=path_col, desc_wid=desc_col))
    for secret in secrets:
        _desc = secret.description if secret.description else 'None'
        if len(_desc) > desc_col:
            _desc = (_desc[:MAX_DESC_WIDTH-2].strip() + '..')
        _tags = json_to_tags(secret.tags) if secret.tags else 'None'
        print(line.format(secret.path, _desc, _tags, path_wid=path_col, desc_wid=desc_col))

def desc_col_length(secrets):
    """calculate width of DESCRIPTION column"""
    temp_secs = secrets.copy()
    temp_secs.append(Secret(path='temp', description='temp'))
    col_w = len(max(temp_secs, key=lambda secret: len(secret.description) if secret.description else 0).description) + 2
    col_w = max([col_w, MIN_DESC_WIDTH])
    return col_w if col_w <= MAX_DESC_WIDTH else MAX_DESC_WIDTH

# export secrets on path in specified format
def print_export(secrets, output='tfvars'):
    """export secret list in desired format"""
    options = {'tfvars': print_tfvars, 'json': print_json, 'csv': print_csv}
    if output in options:
        options[output]([secret.value for secret in secrets])
    else:
        print('secretctl: export format supported include tfvars, json, ')

def print_tfvars(secrets):
    """print key=value pairs to stdout"""
    for secret in secrets:
        for key in secret:
            print(f"{key}={secret[key]}")

def print_json(secrets):
    """print json formatted key: value pairs to stdout"""
    result = ""
    result += "{\n"
    for secret in secrets:
        for key in secret:
            result += f"\"{key}\": \"{secret[key]}\",\n"
    result = result[:-2] + "\n}"
    print(result)

def print_csv(secrets):
    """print csv formatted key,value pairs to stdout"""
    print('Key,Value')
    for secret in secrets:
        for key in secret:
            print(f"{key},{secret[key]}")



"""cli commands"""
import json
from invoke import task

from .validators import validate_path
from .validators import validate_recovery
from .validators import tags_to_json
from .validators import set_secret
from .validators import read_value

from .tuples import create_secret
from .tuples import update_secret
from .tuples import get_secret
from .tuples import tag_secret
from .tuples import untag_secret
from .tuples import list_secrets
from .tuples import delete_secret

from .output import print_read, print_list, print_export


@task(optional=['isjson', 'description', 'tags'])
def create(_ctx, path, value, isjson=False, description=None, tags=None):
    """create path/key value | - [--description STRING] [--tags TAG=STRING, ..] [--isjson]

       Flags:

         --description STRING        Add a description to new secret.

         --tags <tag>=<value>, ...   Include tags with new secret.

         - [--isjson]                Read value from stdin. Include --isjson to validate json string.

                                     $> cat <filename> | secretctl create myapp/dev/public-key -
                                     myapp/dev/public-key created
    """
    secret_kwargs = {}
    secret_kwargs['path'] = validate_path(path)
    secret_kwargs['value'] = read_value(path, value, isjson)
    secret_kwargs['description'] = description
    if tags: secret_kwargs['tags'] = tags_to_json(tags)

    resp = create_secret(**secret_kwargs)
    print(f"{resp.path} created")

@task(optional=['isjson', 'description'])
def update(_ctx, path, value, isjson=False, description=None):
    """update path/key value | - [--description STRING] [--isjson]

       Flags:

         --description STRING        Update description of secret.

         - [--isjson]                Read value from stdin. Include --isjson to validate json string.

                                     $> cat <filename> | secretctl update myapp/dev/public-key -
                                     myapp/dev/public-key updated
    """
    secret_kwargs = {}
    secret_kwargs['path'] = validate_path(path)
    secret_kwargs['value'] = read_value(path, value, isjson)
    secret_kwargs['description'] = description
    resp = update_secret(**secret_kwargs)
    print(f"{resp.path} updated")

@task(optional=['quiet', 'info'])
def read(_ctx, path, quiet=False, info=False):
    """read path/key [--quiet] [--info]

       $>  secretctl read myapp/dev/docker_login
       Path/Key                   Version   Value
       myapp/dev/docker_login     1         mydockerlogin

       Flags:

         --quiet            Return only the secret value. Useful for working with secrets in pipelines.
                            Ex: Set DOCKER_LOGIN = to secret

                            $>  export DOCKER_LOGIN=$(secretctl read myapp/dev/docker_login -q)
                            $>  echo $DOCKER_LOGIN
                            mydockerlogin

         --info             Show description and tags.
    """
    print_read(get_secret(validate_path(path)), quiet=quiet, info=info)


# removed from code coverage pending moto support for mocking secretsmanager tags
@task
def tag(_ctx, path, tags):   # pragma: no cover
    """tag key/value 'TAG=STRING, ..'"""
    secret_kwargs = {}
    secret_kwargs['path'] = validate_path(path)
    secret_kwargs['tags'] = tags_to_json(tags)
    resp = tag_secret(**secret_kwargs)
    print(f"{resp.path} tagged")

# removed from code coverage pending moto support for mocking secretsmanager tags
@task
def untag(_ctx, path, tags):   # pragma: no cover
    """untag path/key 'TAG, ..'"""
    secret_kwargs = {}
    secret_kwargs['path'] = validate_path(path)
    secret_kwargs['tags'] = tags_to_json(tags, novalue=True)
    resp = untag_secret(**secret_kwargs)
    print(f"tags removed from {resp.path}")

#pylint: disable=W0622
@task(optional=['path', 'tags'])
def list(_ctx, path=None, tags=None):
    """list [--path STRING] [--tags STRING]

       Flags:

         --path STRING      Returns the subset of secrets with path STRING.
                            Ex: lists all the secrets for the dev environment of myapp

                            $>  secretctl list -p di/dev/
                            Path/Key                Description                             Tags
                            di/dev/docker_username  access credentials for private regis..  team=di
                            di/dev/docker_password  access credentials for private regis..  team=di
                            di/dev/vault_token      team vault token                        team=di
                            Found 3 secrets.

         --tags STRING      Filter secrets by Tag STRING. Will includes tags with Keys or Values that 'contain' STRG.
                            Ex: lists all the secrets with a Tag containing a team's name, "bravo"

                            $ secretctl list -t bravo
                            Path/Key      Description     Tags
                            app/dev/pem   private key     team=bravo
                            app/qa/pem    private key     team=bravo
    """
    secrets = []
    filter_secrets = list_secrets()
    if path: path = validate_path(path)
    filter_secrets = [sec for sec in filter_secrets if not path or (path and sec['Name'].startswith(path))]
    filter_secrets = [sec for sec in filter_secrets if not tags or ('Tags' in sec and tags in json.dumps(sec['Tags']))]
    for secret in filter_secrets:
        secrets.append(set_secret(secret))

    if len(secrets) >= 1:
        print_list(secrets)
        print(f"Found {len(secrets)} secrets.")
    else:
        print('secretctl: no secrets match filter')

@task(optional=['output'])
def export(_ctx, path, output='tfvars'):
    """export path/ [--output json | csv | tfvars(default)]

       $> secretctl export myapp/dev
       docker_login=username123
       docker_password=passwordABC

       Flags:

         --output [option]  Output format options: tfvars(dotenv), json, csv

                            $>  secretctl export myapp/dev -o json > env_vars.json
                            $>  cat env_vars.json
                            {
                            "docker_login": "username123",
                            "docker_password": "passwordABC"
                            }

    """
    if path: path = validate_path(path)
    secrets = []
    for secret in list_secrets():
        if not path or (path and secret['Name'].startswith(path)):
            secrets.append(get_secret(secret['Name']))

    if len(secrets) >= 1:
        print_export(secrets, output=output)
    else:
        print('secretctl: no secrets match filter')

@task(optional=['recovery'])
def delete(_ctx, path, recovery=7):
    """delete path [--recovery INT]

       $> secretctl delete myapp/dev
       myapp/dev deleted

       Flags:

         --recovery INT  A deleted secret will not be accessible via the cli. The secret still exists
                         with SecretsManager and can be recovered via the Console. SecretsManager schedules
                         the secret for permanent deletion after 7 days by default. You can override the
                         default using this flag, specifying the number of days (max 30) before permanent deletion.

                         $>  secretctl delete myapp/dev -r 14

    """
    secret_kwargs = {}
    secret_kwargs['path'] = validate_path(path)
    secret_kwargs['recovery'] = validate_recovery(recovery)
    resp = delete_secret(**secret_kwargs)
    print(f"{resp['Name']} deleted")



"""validate inputs"""
import re
import sys
import json
from secretctl.tuples import Secret

DELIM = '/'

# validate user path input
def validate_path(path):
    """validate secret path name: characters, delimeter, length"""
    path = path[:-1] if path[-1] == '/' else path
    if not re.match(r"^(?=.{2,256}$)([a-zA-Z0-9_@.-]+\/)*([a-zA-Z0-9_@.-])*[a-zA-Z0-9_@.-]$", path):
        print("secretctl: invalid path/key name. Allowable characters include [A-z0-9_@.-], " \
              "/ for path delimiter, 3 to 256 chars long")
        sys.exit(1)
    return path

# validate user recovery days input
def validate_recovery(recovery):
    """validate recovery days: integer, length"""
    if int(recovery) in range(7, 30):
        return recovery
    print('secretctl: invalid recovery days. Must be integer between 7 and 30')
    sys.exit(1)

# convert cli provided tags into key:value json
def tags_to_json(tags, novalue=False):
    """validate supplied tags and convert to secretsmanager tags json"""
    options = {True: tags_without_values, False: tags_with_values}
    return options[novalue](tags)

def tags_with_values(tags):
    """validate supplied tags-with-values and convert to secretsmanager tags json"""
    tag_list = []
    for tag in tags.split(","):
        if not re.match(r"^(([a-zA-Z0-9\/\+\:_@.-]{3,127})\s*[=]\s*([a-zA-Z0-9\/\+\:_@.-]{1,255}))$", tag.strip()):
            print("secretctl: invalid tags list. Supply tags as \"tag1=value1, tag2=value2, ...\"")
            sys.exit(1)
        tag_list.append({"Key": tag.split("=")[0].strip(), "Value": tag.split("=")[1].strip()})
    return tag_list

def tags_without_values(tags):
    """validate supplied tags-without-values and convert to secretsmanager tags json"""
    tag_list = []
    for tag in tags.split(","):
        if not re.match(r"^([a-zA-Z0-9\/\+\:_@.-]{3,127})$", tag.strip()):
            print("secretctl: invalid remove tags list. Supply tags as \"tag1, tag2, ...\"")
            sys.exit(1)
        tag_list.append(tag.strip())
    return tag_list

# convert json Tags to simple cli formatted
def json_to_tags(tags):
    """format json Tags as simple tag list"""
    tag_list = ""
    for index, _ in enumerate(tags):
        if index == 0:
            tag_list += (tags[index]['Key'] + '=' + tags[index]['Value'])
        else:
            tag_list += (', ' + tags[index]['Key'] + '=' + tags[index]['Value'])
    return tag_list

def read_value(path, value, isjson=False):
    """parse supplied value or response from sys.stdin.read (pipe)"""
    resp = sys.stdin.read() if value == '-' else value
    if isjson:
        try:
            json.loads(resp)
        except ValueError as e:
            print('secretctl: invalid json %s' % e)
            sys.exit(1)
    else:
        resp = json.dumps({path.split(DELIM)[-1]: value})
    return resp

def set_secret(secret):
    """return Secret from list_secret element"""
    secret_kwargs = {}
    secret_kwargs['path'] = secret['Name']
    if 'Description' in secret:
        secret_kwargs['description'] = secret['Description']
    if 'Tags' in secret:
        secret_kwargs['tags'] = secret['Tags']
    return Secret(**secret_kwargs)



import os
from setuptools import setup, find_packages
from _version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements=[
    'invoke==1.3.0',
    'boto3==1.9.226',
    'pyyaml==5.1.2',
    'requests==2.22.0',
    'colorama==0.4.1'
]

setup(
    name="secretctl",
    author="Nic Cheneweth",
    author_email="nic.cheneweth@thoughtworks.com",
    description="Command line tool for working with aws secrets manager",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ncheneweth/secretctl",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        'Development Status :: 3 - Alpha',
    ],
    python_requires='>=3.5',
    version=__version__,
    install_requires=requirements,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    entry_points={
        'console_scripts': ['secretctl = secretctl.main:program.run']
    }
)
